from . import instruments
import astropy.coordinates as coord
import astropy.units as u
import numpy as np
from dataclasses import dataclass, field
import logging
from ..utils.utilities import logger_default

logger = logging.getLogger(__name__)
logger = logger_default(logger) 


@dataclass
class Telescope():
    name: str
    location: coord.earth.EarthLocation
    instruments: list = field(default_factory=list)
    diameter: u.Quantity = 0 *u.m,
    operational: bool = True,
    variable_name: str = ''


# Calar Alto Observatory (CARMENES)
CalarAlto = Telescope(
    name='Calar Alto Observatory',
    location=coord.EarthLocation.of_site('Observatorio de Calar Alto'),
    instruments=[instruments.CARMENES],
    diameter=3.5 * u.m
)

# La Silla Observatory, Swiss Euler Telescope (CORALIE)
SwissEuler = Telescope(
    name='Swiss Euler Telescope',
    location=coord.EarthLocation.of_site('La Silla Observatory'),
    instruments=[instruments.CORALIE],
    diameter=1.2 * u.m
)

# La Silla Observatory, ESO 3.6-meter Telescope (HARPS, NIRPS)
LaSilla3_6m = Telescope(
    name='ESO 3.6m Telescope (La Silla)',
    location=coord.EarthLocation.of_site('La Silla Observatory (ESO)'),
    instruments=[instruments.HARPS, instruments.NIRPS],
    diameter=3.6 * u.m
)

# Paranal Observatory, VLT Unit Telescope (ESPRESSO, UVES)
VLT = Telescope(
    name='Very Large Telescope (VLT)',
    location=coord.EarthLocation.of_site('Paranal'),
    instruments=[instruments.ESPRESSO, instruments.UVES],
    diameter=8.2 * u.m
)

# Paranal Observatory, VLT 4-UT (ESPRESSO 4UT)
VLT_4UT = Telescope(
    name='Very Large Telescope (VLT) 4-UT',
    location=coord.EarthLocation.of_site('Paranal'),
    instruments=[instruments.ESPRESSO_4UT],
    diameter=2 * 8.2 * u.m  # Combined light from all 4 UTs
)

# Lowell Observatory, Lowell Discovery Telescope (EXPRES)
LowellDiscovery = Telescope(
    name='Lowell Discovery Telescope',
    location=coord.EarthLocation.of_site('Lowell Observatory'),
    instruments=[instruments.EXPRES],
    diameter=4.3 * u.m
)

# Roque de los Muchachos Observatory, Telescopio Nazionale Galileo (GIANO, HARPS-N)
TNG = Telescope(
    name='Telescopio Nazionale Galileo (TNG)',
    location=coord.EarthLocation.of_site('Roque de los Muchachos'),
    instruments=[instruments.GIANO, instruments.HARPS_N],
    diameter=3.58 * u.m
)

# Mauna Kea Observatory, Gemini North Telescope (MAROON-X)
GeminiNorth = Telescope(
    name='Gemini North Telescope',
    location=coord.EarthLocation.of_site('gemini_north'),
    instruments=[instruments.MAROON_X],
    diameter=8.1 * u.m
)

# Mauna Kea Observatory, Canada-France-Hawaii Telescope (SPIRou)
CFHT = Telescope(
    name='Canada-France-Hawaii Telescope (CFHT)',
    location=coord.EarthLocation.of_site('Canada-France-Hawaii Telescope'),
    instruments=[instruments.SPIROU],
    diameter=3.6 * u.m
)

# Haute-Provence Observatory (SOPHIE)
HauteProvence = Telescope(
    name='Haute-Provence Observatory',
    location=coord.EarthLocation.of_site('Observatoire de Haute Provence'),
    instruments=[instruments.SOPHIE],
    diameter=1.93 * u.m
)


def print_all_telescopes(instruments: str = 'all') -> None:
    if instruments != 'all':
        raise NotImplementedError
    
    telescopes_list = [(telescope_name, telescope) for telescope_name, telescope in globals().items() if isinstance(telescope, Telescope)]
    telescopes_list.sort(reverse= True, key=lambda telescopes_list: telescopes_list[1].diameter.to(u.m).value)
    
    logger.print('Printing available telescopes:')
    logger.info('Use the first name to access the telescope object.')
    logger.info('    e.g., to access the VLT telescope, use "VLT" instead of "Very Large Telescope (VLT)"')
    logger.print('='*25)
    for (telescope_name, telescope) in telescopes_list:
        logger.print(f"{telescope_name} : {telescope.name} | {telescope.diameter} | Operational: {bool(telescope.operational)}")
        logger.print(f"    {[instrument.name for instrument in telescope.instruments]}")
        
        
if __name__ == '__main__':
    logger.warning('='*25)
    logger.warning('Debugging mode: Telescopes module')
    logger.warning('='*25)
    print_all_telescopes()
    from . import instruments as inst
    inst.print_all_spectrographs()
    
    print([mode for mode in inst.ESPRESSO.modes if mode.exposure_time_calculator is not None])
    
    for mode in [mode for mode in inst.ESPRESSO.modes if mode.exposure_time_calculator is not None]:
        mode.exposure_time_calculator.open_all_scenarios(stellar_temperature= 5500)
    
    logger.warning('='*25)
    logger.warning('End of debugging mode: Telescopes module')
    logger.warning('='*25)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Aug 22 10:35:50 2023

@author: chamaeleontis
"""
#%% Importing libraries
import sys
import pandas as pd
# from PTO.transits.NASA_exo import NASA_Exoplanets_archive, logger
import datetime
import astropy.units as u
import numpy as np
import astropy.time as astime
import astropy.coordinates as coord
import matplotlib.pyplot as plt
import seaborn as sns
from dataclasses import dataclass
# from PTO.PTO.transits.observatories import Observatories
import re
import matplotlib.dates as md 
import os

#%%
def define_baseline(time_t14:float) -> float:
    """
    Define baseline based on input time.

    Parameters
    ----------
    time_t14 : float
        Transit duration length in hours.

    Returns
    -------
    baseline_length : float
        Length of baseline.

    """
    baseline_length = np.min([3, .75*time_t14])
    baseline_length = np.max([baseline_length, 2])
    return baseline_length

#%% Class definitions
#%% Planet transits
@dataclass(init=True, repr=False, eq=False, order=False, unsafe_hash=False, frozen=False)
class _Planet_transits:
    """Class holding all the transit windows and parameters for convenience."""
    name: str
    transit_windows: list
    parameters: pd.core.series.Series


@dataclass(init=True, repr=False, eq=False, order=False, unsafe_hash=False, frozen=False)
class _TimeArray:
    """Class holding timings of the transit."""
    def _define_times(self,
                      Night,
                      Planet_transit)->None:
        """Define basic times and time arrays for the plot."""
        self.midnight = astime.Time(round(Night.value) - 0.5,
                                    format=Night.format)
        
        self.time_array = np.linspace(-12,12,24*60+1)*u.hour + self.midnight
        self.T1, self.T4 = ((Night - Planet_transit.parameters['pl_trandur']*u.hour/2) ,
                            (Night + Planet_transit.parameters['pl_trandur']*u.hour/2))
        return None
    
    def _define_sunset_and_sunrise(self,
                                   sunset,
                                   sunrise)->None:
        self.Sunset = sunset
        self.Sunrise = sunrise
        

@dataclass(init=True, repr=False, eq=False, order=False, unsafe_hash=False, frozen=False)
class _Indices:
    """Class holding indices of various arrays."""
    def _define_indices(self,
                        TimeArray,
                        AltitudeArray)->None:
        """Define the indices of transit, ingress, egress and out-of-transit."""
        self.Transit = np.where(
            np.logical_and(
                TimeArray.time_array > TimeArray.T1,
                TimeArray.time_array < TimeArray.T4
                )
            )
        self.Out_of_transit = np.where(
            np.logical_or(
                TimeArray.time_array < TimeArray.T1,
                TimeArray.time_array > TimeArray.T4
                )
            )
        self.Ingress = np.where(
                TimeArray.time_array < TimeArray.T1,
            )
        self.Egress = np.where(
                TimeArray.time_array > TimeArray.T4
            )
        self.Observable = np.where(
            np.logical_and(
                AltitudeArray.target.secz < 2.2,
                AltitudeArray.target.secz >= 1
                )
            )

class _Visibility:
    """Class holding the indices when baseline, target, transit, ingress and egress is visible."""
    def _define_visibility(self,
                           NightDefiningIndices,
                           Indices
                           )->None:
        """Define the visibility depending on full/twilight mode and Indices of transit."""
        self.Visibility = np.intersect1d(Indices.Observable, NightDefiningIndices)
        self.Transit = np.intersect1d(self.Visibility,
                                      Indices.Transit)
        self.Baseline = np.intersect1d(self.Visibility,
                                       Indices.Out_of_transit)
        self.Ingress = np.intersect1d(self.Visibility,
                                      Indices.Ingress)
        self.Egress = np.intersect1d(self.Visibility,
                                      Indices.Egress)
        return

class _Observations:
    """Class holding the indices for the observations."""
    def _define_observations(self,
                             Visibility: _Visibility,
                             baseline_length: float
                             ):
        baseline_length = int(round(baseline_length*60/2))
        
        if len(Visibility.Ingress) > baseline_length:
            self.Ingress = Visibility.Ingress[-baseline_length:]
            missing_length_ingress = 0
        else:
            self.Ingress = Visibility.Ingress
            missing_length_ingress = baseline_length - len(Visibility.Ingress)
            
        if len(Visibility.Egress) > baseline_length:
            self.Egress = Visibility.Egress[:baseline_length]
            missing_length_egress = 0
        else:
            self.Egress = Visibility.Egress
            missing_length_egress = baseline_length - len(Visibility.Egress)
        
        if missing_length_ingress != 0 and missing_length_egress != 0:
            self.Ingress, self.Egress = Visibility.Ingress, Visibility.Egress
        elif missing_length_ingress != 0:
            if len(Visibility.Egress) > (baseline_length+missing_length_egress):
                self.Egress = Visibility.Egress[:(baseline_length+missing_length_ingress)]
            else:
                self.Egress = Visibility.Egress
        elif missing_length_egress != 0:
            if len(Visibility.Ingress) > (baseline_length+missing_length_ingress):
                self.Ingress = Visibility.Ingress[-(baseline_length+missing_length_egress):]
            else:
                self.Ingress = Visibility.Ingress
        
        self.Transit = Visibility.Transit
        self.complete = np.concatenate([self.Ingress,self.Transit, self.Egress])

        return


@dataclass(init=True, repr=False, eq=False, order=False, unsafe_hash=False, frozen=False)
class _AltitudeArray:
    """Class holding alitudes of Sun, Moon and target."""
    
    def _define_targets(self,
                        TimeArray,
                        Location:str,
                        Planet_transit,
                        )->None:
        """Define the altitude azimuth frame and calculates target, Moon and Sun position in it."""
        self.altitude_azimuth_frame = coord.AltAz(
            obstime = TimeArray.time_array,
            location = Observatories[Location].value
            )
        self.Sun = coord.get_sun(TimeArray.time_array
                                 ).transform_to(
                                     self.altitude_azimuth_frame)
        self.Moon = coord.get_body("moon",
                                TimeArray.time_array,
                                location = Observatories[Location].value
                                ).transform_to(
                                    self.altitude_azimuth_frame)
        self.target = coord.SkyCoord(
            ra = coord.Angle(Planet_transit.parameters['ra'] * u.deg),
            dec = coord.Angle(Planet_transit.parameters['dec'] * u.deg),
            frame='icrs'
            ).transform_to(
                self.altitude_azimuth_frame)
        self.MoonSeparation = np.min(self.Moon.separation(self.target))
        
        ind = np.argwhere((self.Sun.alt.value < 0) * self.Sun.alt.value != 0)
        TimeArray._define_sunset_and_sunrise(TimeArray.time_array[ind[0]].to_value(format= 'datetime')[0],
                                             TimeArray.time_array[ind[-1]].to_value(format= 'datetime')[0])
        
        return None

#%% Flags transit
@dataclass(init=True, repr=False, eq=False, order=False, unsafe_hash=False, frozen=False)
class _Flags_transit:
    """Flags for given transit opportunity in terms of visibility."""
    def check(self,
              Visibility: _Visibility,
              parameters: pd.DataFrame,
              baseline_length: float,
              partial : bool | float = False
              ):
        """
        Check the values of flags.

        Parameters
        ----------
        Visibility : _Visibility
            Visibility of target's transit. Can be full night or twilight-included version.
        parameters : pd.DataFrame
            Planetary parameters.
        baseline_length : float
            Baseline length.

        Returns
        -------
        None.

        """
        if partial == False:
            partial = 1
        
        # Visibility of transit
        if (len(Visibility.Transit) > (parameters['pl_trandur'] * 60)-1):
            self.transit = True
            self.transit_coverage = 1
        else:
            self.transit = False
            self.transit_coverage = ((len(Visibility.Transit))/ (parameters['pl_trandur'] * 60))
            
            if self.transit_coverage > partial:
                self.transit = True
            
        
        # Baseline visibility
        self.baseline = (len(Visibility.Baseline) > baseline_length * 60)
        self.baseline_ingress = (len(Visibility.Ingress) > baseline_length/2*60)
        self.baseline_egress = (len(Visibility.Egress) > baseline_length/2*60)
        
        self.baseline_coverage = np.min([( len(Visibility.Baseline)) / (baseline_length * 60),1])
        self.baseline_ingress_coverage = np.min([( len(Visibility.Ingress) / (baseline_length/2 * 60)),1])
        self.baseline_egress_coverage = np.min([(len(Visibility.Egress) / (baseline_length/2 * 60) ),1])
        
        if self.baseline_coverage > partial:
            self.baseline = True
        return None
#%% Flags for a given window
@dataclass(init=True, repr=False, eq=False, order=False, unsafe_hash=False, frozen=False)
class _Flags_window:
    """Flags for given transit window in terms of feasibility."""
    moon_angle_warning: bool = False # 30 deg < separation < 45 deg
    moon_angle_critical: bool = False # separation < 30 deg
    
    def check(self,
              FlagsTransitFull: _Flags_transit,
              FlagsTransitTwilight: _Flags_transit,
              MoonSeparation: float
              ):
        """
        Check the flags values.

        Parameters
        ----------
        FlagsTransitFull : _Flags_transit
            Flags for full night observation.
        FlagsTransitTwilight : _Flags_transit
            Flags for twilight-included observation.
        MoonSeparation : float
            Minimal separation between target and Moon.

        Returns
        -------
        None.

        """
        
        if FlagsTransitFull.transit and FlagsTransitFull.baseline:
            self.visible = True
        else:
            self.visible = False
            
        if FlagsTransitTwilight.transit and FlagsTransitTwilight.baseline:
            self.visible_twilight = True
        else:
            self.visible_twilight = False
        
        if MoonSeparation <30*u.deg:
            self.moon_angle_critical = True
            self.moon_angle_warning = False
        elif MoonSeparation < 45 * u.deg:
            self.moon_angle_critical = False
            self.moon_angle_warning = True
        else:
            self.moon_angle_critical = False
            self.moon_angle_warning = False
        return
#%% Transit plots
@dataclass(init=True, repr=False, eq=False, order=False, unsafe_hash=False, frozen=False)
class Plot_transit:
    """Class holding the settings for a single transit window plot."""
    
    Location: str
    Planet_transit: _Planet_transits
    Night: astime.Time
    Uncertainty: u.quantity.Quantity
    partial: bool | float
    TimeArray: _TimeArray = _TimeArray()
    AltitudeArray: _AltitudeArray = _AltitudeArray()
    VisibilityFull: _Visibility = _Visibility()
    VisibilityTwilight: _Visibility = _Visibility()
    Indices: _Indices = _Indices()
    FlagsFull: _Flags_transit = _Flags_transit()
    FlagsTwilight: _Flags_transit = _Flags_transit()
    FlagsWindow: _Flags_window = _Flags_window()
    ObservationsFull: _Observations = _Observations()
    ObservationsTwilight: _Observations = _Observations()
    def __post_init__(self):
        """
        Automatic setup of the plot and quality estimate.

        Returns
        -------
        None.

        """
        self._define_arrays()        
        self._calculate_baseline_observations()
        self._define_observability()
        self._define_flags()
        self._estimate_quality_of_transit_window()
        
        if self.quality > 0:
            self._log_transit_info()
            self._create_plot()
        
        return
    
    def _create_plot(self):
        """
        Create transit window plot.

        Returns
        -------
        None.

        """
        with plt.ioff():
            fig, ax = plt.subplots(1,figsize=(18,12))
            self._plot_basic_setting(fig, ax)
            self._plot_twilight(ax)
            self._plot_Moon(ax)
            self._plot_target(ax)
            self._plot_observation(ax)
            self._write_system_parameters(ax)
            
            self.fig, self.ax = fig, ax
        return
    
    def _log_transit_info(self):
        """
        Log transit information and quality.

        Returns
        -------
        None.

        """
        logger.info('Transit during night: ' + str(self.TimeArray.midnight.to_value('datetime').strftime('%Y%m%d')) + ' is of quality: '+ str(self.quality) +' with uncertainty of {sigma:.2f} min'.format(sigma = self.Uncertainty.to(u.min).value))
        if self.quality != 1:
            logger.warning('This is a degraded transit because:')
            if self.FlagsWindow.moon_angle_critical:
                logger.warning('    Moon separation angle too low')
            if self.FlagsWindow.moon_angle_warning:
                logger.warning('    Moon separation angle warning (30 deg < angle < 45 deg)')
            if (not(self.FlagsWindow.visible) and
                self.FlagsWindow.visible_twilight):
                logger.warning('    Transit window is heavily visible inside the twilight.')
            if self.FlagsFull.baseline_ingress_coverage < 0.9:
                logger.warning(f'    Ingress baseline is not full with coverage {self.FlagsFull.baseline_ingress_coverage:.0%} during full night')
                logger.info(f'    Ingress baseline including twilight coverage:{self.FlagsTwilight.baseline_ingress_coverage:.0%}')
            if self.FlagsFull.baseline_egress_coverage < 0.9:
                logger.warning(f'    Egress baseline is not full with coverage {self.FlagsFull.baseline_egress_coverage:.0%} during full night')
                logger.info(f'    Egress baseline including twilight coverage: {self.FlagsTwilight.baseline_egress_coverage:.0%}')
            if self.FlagsFull.transit_coverage < 0.9:
                logger.warning(f'    Transit event is not full with coverage {self.FlagsFull.transit_coverage:.0%} during full night')
                logger.info(f'    Transit event including twilight coverage: {self.FlagsTwilight.transit_coverage:.0%}')
        
    def _plot_target(self,
                     ax:plt.Axes):
        """
        Plot the target, highlighting the transit area.

        Parameters
        ----------
        ax : plt.Axes
            Artist on which to plot.

        Returns
        -------
        None.

        """
        ax.plot(self.TimeArray.time_array.to_value(format= 'datetime'),
                self.AltitudeArray.target.alt.value,
                ls = '-',
                label='Target',
                linewidth=4,
                color='red',
                )
        ax.plot(self.TimeArray.time_array[self.Indices.Transit].to_value(format= 'datetime'),
                self.AltitudeArray.target[self.Indices.Transit].alt.value,
                ls = '-',
                label='Target',
                linewidth=10,
                color='red',
                )
        
        ax.scatter(
            self.TimeArray.time_array[self.Indices.Transit[0]].to_value(format= 'datetime'),
            self.AltitudeArray.target[self.Indices.Transit[0]].alt.value,
            color = 'white',
            )
        ax.scatter(
            self.TimeArray.time_array[self.Indices.Transit[-1]].to_value(format= 'datetime'),
            self.AltitudeArray.target[self.Indices.Transit[-1]].alt.value,
            color = 'white',
            )
        
        return
    
    def _plot_Moon(self,
                   ax:plt.Axes):
        """
        Overplot the Moon position.

        Parameters
        ----------
        ax : plt.Axes
            Artist on which to plot on.

        Returns
        -------
        None.

        """
        ax.plot(self.TimeArray.time_array.to_value(format= 'datetime'),
                self.AltitudeArray.Moon.alt.value,
                ls = '-',
                label='Moon',
                linewidth=2,
                color='yellow',
                alpha=0.7
                )
        return
    
    def _plot_twilight(self,
                       ax:plt.Axes):
        """
        Overplot the twilight area.

        Parameters
        ----------
        ax : plt.Axes
            Artist on which to plot on.

        Returns
        -------
        None.

        """
        ax.fill_between(self.TimeArray.time_array.to_value(format= 'datetime'),
                        0, 90,
                        self.AltitudeArray.Sun.alt.value < 0,
                        color='blue',
                        alpha = 0.3
                        )
        ax.fill_between(self.TimeArray.time_array.to_value(format= 'datetime'),
                        0, 90,
                        self.AltitudeArray.Sun.alt.value < -18,
                        color ='blue',
                        alpha= 0.5
                        )
        return

    def _plot_observation(self,
                          ax):
        
        
        #CLEANME
        if len(self.ObservationsFull.complete) != 0:
            ax.fill_between(self.TimeArray.time_array[self.ObservationsFull.complete[0]:
                                            self.ObservationsFull.complete[-1]].to_value(
                                                format= 'datetime'),
                            0, 10,
                            color='lime',
                            alpha = 1
                            )
            ax.text((self.TimeArray.time_array[self.ObservationsFull.complete[0]]+30*u.min).to_value(
                format= 'datetime'),
                    5,
                    'Observations (no twilight): ' + 
                    self.TimeArray.time_array[self.ObservationsFull.complete[0]].to_value(
                        format= 'datetime').strftime('%H:%M') + ' - ' +
                    self.TimeArray.time_array[self.ObservationsFull.complete[-1]].to_value(
                        format= 'datetime').strftime('%H:%M'),
                    fontsize = 10,
                    )
            ax.text(self.TimeArray.midnight.to_value(
                format= 'datetime'),
                    -10,
                    'Observations (no twilight): ' + 
                    self.TimeArray.time_array[self.ObservationsFull.complete[0]].to_value(
                        format= 'datetime').strftime('%H:%M') + ' - ' +
                    self.TimeArray.time_array[self.ObservationsFull.complete[-1]].to_value(
                        format= 'datetime').strftime('%H:%M'),
                    fontsize = 15,
                    bbox=dict(
                        facecolor='green',
                        edgecolor='black',
                        boxstyle='round,pad=1'
                        )
                    )
            
            self.full_observation_time = [
                self.TimeArray.time_array[self.ObservationsFull.complete[0]].to_value(format= 'datetime'),
                self.TimeArray.time_array[self.ObservationsFull.complete[-1]].to_value(format= 'datetime')
                ]
        
        self.twilight_observation_time = None
        
        if ((self.FlagsFull.baseline_ingress) or # Ingress baseline missing
            (self.FlagsFull.baseline_egress) or # Egress baseline missing
            (not(self.FlagsWindow.visible) and self.FlagsWindow.visible_twilight)): # Transit into twilight
            
            
            
            ax.fill_between(self.TimeArray.time_array[self.ObservationsTwilight.complete[0]:
                                            self.ObservationsTwilight.complete[-1]].to_value(
                                                format= 'datetime'),
                            10, 20,
                            color='goldenrod',
                            alpha = 1
                            )

            ax.text((self.TimeArray.time_array[self.ObservationsTwilight.complete[0]]+30*u.min).to_value(
                format= 'datetime'),
                    15,
                    'Observation (with twilight): ' +
                    self.TimeArray.time_array[self.ObservationsTwilight.complete[0]].to_value(
                             format= 'datetime').strftime('%H:%M') + ' - ' +
                    self.TimeArray.time_array[self.ObservationsTwilight.complete[-1]].to_value(
                             format= 'datetime').strftime('%H:%M'),
                    fontsize = 10,
                    )
            
            ax.text(self.TimeArray.midnight.to_value(
                format= 'datetime'),
                    -20,
                    'Observation (with twilight): ' +
                    self.TimeArray.time_array[self.ObservationsTwilight.complete[0]].to_value(
                             format= 'datetime').strftime('%H:%M') + ' - ' +
                    self.TimeArray.time_array[self.ObservationsTwilight.complete[-1]].to_value(
                             format= 'datetime').strftime('%H:%M'),
                    fontsize = 15,
                    bbox=dict(
                        facecolor='orange',
                        edgecolor='black',
                        boxstyle='round,pad=1'
                        )
                    )
            
            ax.axvline(self.TimeArray.time_array[self.ObservationsTwilight.complete[0]].to_value(format= 'datetime'), ls='--', lw=2, color='orange')
            ax.axvline(self.TimeArray.time_array[self.ObservationsTwilight.complete[-1]].to_value(format= 'datetime'), ls='--', lw=2, color='orange')
            self.twilight_observation_time = [
                self.TimeArray.time_array[self.ObservationsTwilight.complete[0]].to_value(format= 'datetime'),
                self.TimeArray.time_array[self.ObservationsTwilight.complete[-1]].to_value(format= 'datetime')
                ]
        
        ax.axvline(self.TimeArray.T1.to_value(format= 'datetime'), ls='--', lw=2, color='darkred')
        ax.axvline(self.TimeArray.T4.to_value(format= 'datetime'), ls='--', lw=2, color='darkred')
        if len(self.ObservationsFull.complete) != 0:
            ax.axvline(self.TimeArray.time_array[self.ObservationsFull.complete[0]].to_value(format= 'datetime'), ls='--', lw=2, color='lime')
            ax.axvline(self.TimeArray.time_array[self.ObservationsFull.complete[-1]].to_value(format= 'datetime'), ls='--', lw=2, color='lime')
        
        
        
        return
    
    def _plot_basic_setting(self,
                            fig: plt.Figure,
                            ax:plt.Axes):
        """
        Plot basic settings regardless of what planet is plotted.

        Parameters
        ----------
        fig : plt.Figure
            Figure on which to plot on.
        ax : plt.Axes
            Artist on which to plot on.

        Returns
        -------
        None.

        """
        Night = (self.TimeArray.midnight-1*u.day).to_value('datetime')
        ax.set_title('Transit of '+ str(self.Planet_transit.parameters['pl_name'])+
                     ' on night: '+str(Night.strftime('%Y%m%d')) + 
                     ' at Paranal with precision on Tc of: ' + 
                     '{sigma:.2f} min'.format(sigma = self.Uncertainty.to(u.min).value) +
                     '; Quality of transit:' + str(self.quality),
                     fontsize = 18
                     )

        ax.axhline(
            np.arcsin(1/2.2)* 180/np.pi,
            ls='--', lw=2,
            color= 'darkred'
            )
        ax.grid(color='black', linestyle='dashed', linewidth=0.3, alpha=1.0)
        xfmt = md.DateFormatter('%d/%m  %H')
        ax.xaxis.set_major_formatter(xfmt)
        fig.autofmt_xdate()
        hours = md.HourLocator()
        ax.xaxis.set_major_locator(hours)
        
        ax.set_ylim(0,90)
        ax2 = ax.twinx()
        ax.set_ylabel('Altitude [deg]', color='k')
        ax2.set_ylabel('Airmass', color='k')
        ax2.tick_params('y', colors='k')
        ax2.set_ylim([0,90])
        ax2.set_yticklabels(['',5.76,2.92,2.00,1.56,1.31,1.15,1.06,1.02,1.00]) 
        
        ax.set_xlim(self.TimeArray.Sunset,
                    self.TimeArray.Sunrise)
        
        
        return
    
    def _write_system_parameters(self,
                                 ax: plt.Axes):
        """
        Write the system parameters for given system, including Right Ascension, Declination,V magnitude, Moon angle separation, Transit length, Period, ingress and egress timings.

        Parameters
        ----------
        ax : plt.Axes
            DESCRIPTION.

        Returns
        -------
        None.

        """
        ax.text(self.TimeArray.Sunset,
                97,
                ('RA: {ra:.2f}'.format(ra=self.Planet_transit.parameters['ra']) + 
                '; DEC: {dec:.2f}'.format(dec=self.Planet_transit.parameters['dec']) + \
                '; V = {V_mag:.3f}'.format(V_mag = self.Planet_transit.parameters['sy_vmag']) + \
                '; Moon = {moon:.0f} deg'.format(moon = self.AltitudeArray.MoonSeparation.value) + \
                '; $T_{14}$' + ' = {t14:.3f} h'.format(t14 = self.Planet_transit.parameters['pl_trandur']) + \
                '; P = {p:.5f} d'.format(p=self.Planet_transit.parameters['pl_orbper'])
                ),
                style='italic',size=12
                )
        
        ax.text(self.TimeArray.Sunset,
                -23,
                "ingress=" + self.TimeArray.T1.to_value('datetime').strftime("%H:%M"),
                style='italic',
                bbox=dict(facecolor='none', edgecolor='black', pad=7)
                )
        
        ax.text(self.TimeArray.Sunrise,
                -23,
                "egress=" + self.TimeArray.T4.to_value('datetime').strftime("%H:%M"),
                style='italic',
                bbox=dict(facecolor='none', edgecolor='black', pad=7)
                )
        return
    
    def _calculate_baseline_observations(self):
        """Calculate how long baseline we need."""
        self.baseline_length = define_baseline(self.Planet_transit.parameters['pl_trandur'])
        return
    
    def _define_arrays(self):
        """Define the arrays for the plot."""
        self.TimeArray._define_times(Night = self.Night,
                                    Planet_transit = self.Planet_transit,)
        
        self.AltitudeArray._define_targets(self.TimeArray,
                                           self.Location,
                                           self.Planet_transit
                                           )
        
        self.Indices._define_indices(self.TimeArray,
                                     self.AltitudeArray)
        full_defining_indices = np.argwhere(self.AltitudeArray.Sun.alt.value < -18)
        
        self.VisibilityFull._define_visibility(NightDefiningIndices = full_defining_indices,
                                               Indices = self.Indices)
        
        twilight_defining_indices = np.argwhere(
            (self.AltitudeArray.Sun.alt.value < 0) * self.AltitudeArray.Sun.alt.value != 0
            )
        self.VisibilityTwilight._define_visibility(NightDefiningIndices = twilight_defining_indices,
                                                   Indices = self.Indices)
        
        return
    
    def _define_flags(self):
        """
        Define the flags to estimate transit window quality.

        Returns
        -------
        None.

        """
        self.FlagsFull.check(Visibility = self.VisibilityFull,
                             parameters = self.Planet_transit.parameters,
                             baseline_length = self.baseline_length,
                             partial= self.partial
                             )
        self.FlagsTwilight.check(Visibility = self.VisibilityTwilight,
                                 parameters = self.Planet_transit.parameters,
                                 baseline_length = self.baseline_length,
                                 partial= self.partial
                                 )
        self.FlagsWindow.check(
            FlagsTransitFull= self.FlagsFull,
            FlagsTransitTwilight= self.FlagsTwilight,
            MoonSeparation= self.AltitudeArray.MoonSeparation)
        
        return
    
    def _define_observability(self):
        
        self.ObservationsFull._define_observations(
            Visibility= self.VisibilityFull,
            baseline_length= self.baseline_length
            )
        self.ObservationsTwilight._define_observations(
            Visibility= self.VisibilityTwilight,
            baseline_length= self.baseline_length
            )
        return
    
    def _estimate_quality_of_transit_window(self):
        """
        Estimate quality of transit window.
        
        The flagging works like this:
        If not visible (even in twilight) quality = -9999 (ignored completely)
        
        If not visible during full night quality = 5
        
        Else quality = 1
        
        Afterwards, multiple increments are added in case of bad additional condition.
        

        Returns
        -------
        None.

        """
        
        if not(self.FlagsWindow.visible):
            if not(self.FlagsWindow.visible_twilight):
                self.quality = -9999
                return
            else:
                self.quality = 5
        else:
            self.quality = 1
        
        if self.FlagsWindow.moon_angle_critical:
            self.quality += 10
        else:
            if self.FlagsWindow.moon_angle_warning:
                self.quality += 1
        
        if not(self.FlagsFull.baseline):
            if self.FlagsFull.baseline_coverage > 0.8:
                self.quality += 1
            else:
                if self.FlagsTwilight.baseline_coverage != 1:
                    self.quality += 2
                else:
                    self.quality += 1
        
        if self.FlagsFull.baseline == True:
            if ((self.FlagsFull.baseline_ingress_coverage < 1) or
                (self.FlagsFull.baseline_egress_coverage < 1)):
                if ((self.FlagsTwilight.baseline_ingress_coverage < 1) or
                    (self.FlagsTwilight.baseline_egress_coverage < 1)):
                    self.quality += 2
                else:
                    self.quality += 1
        elif (self.FlagsFull.baseline_coverage > 0.8):
            self.quality += 2
        else:
            self.quality += 4
        
        
        return

    def save_transit_window(self,
                            location:str
                            ):
        """
        Save the transit window plot at given location.

        Parameters
        ----------
        location : str
            Filepath to the saved figure.

        Returns
        -------
        None.

        """
        self.filename = location
        self.fig.savefig(location)
        plt.close('all')
        return
    
    def output_observability(self):
        
        # Full observability
        # Planned observability
        
        return

#%%
class Transit_windows:
    """Class holding all transit windows calculated."""
    def __init__(
            self,
            database: NASA_Exoplanets_archive,
            semester_date: astime.Time,
            save_directory: str,
            ):
        """
        Initialize the class Transit windows. 

        Parameters
        ----------
        database : NASA_Exoplanets_archive
            Which database we take data from.
        semester_date : astime.Time
            Starting and ending date of the observations.
        save_directory : str
            Save directory for the figures.

        Returns
        -------
        None.

        """
        self.planets = []
        
        filtered_table = database.filtered_table_all
        filtered_table = self._check_parameters_in_table(filtered_table)
        
        filtered_table_composite = database.filtered_table_composite
        filtered_table_composite = self._check_parameters_in_table(filtered_table_composite,
                                                                   True)
        filtered_table_all = self._add_transit_length_if_missing(filtered_table_all= filtered_table,
                                                                 filtered_table_composite= filtered_table_composite)
        
        
        self.calculate_transit_windows(
            exoplanet_table = filtered_table,
            date = semester_date
            )
        
        self.calculate_transit_windows(
            exoplanet_table = filtered_table_composite,
            date = semester_date
            )
        
        self.save_directory = save_directory
        
        return
    
    def _check_parameters_in_table(self,
                                   filtered_table:pd.DataFrame,
                                   composite:bool = False) -> pd.DataFrame:
        """
        Filter out references without full ephemeris parameters.

        Parameters
        ----------
        filtered_table : pd.DataFrame
            Filtered table from the NASA table.
        composite : bool, optional
            Whether we use composite/complete tables. The default is False.

        Returns
        -------
        filtered_table : pd.DataFrame
            Filtered table.

        """
        original_length = len(filtered_table)
        condition_composite = (filtered_table['pl_orbper'].notna() &
                               filtered_table['pl_orbpererr1'].notna() &
                               filtered_table['pl_orbpererr2'].notna() &
                               filtered_table['pl_tranmid'].notna() &
                               filtered_table['pl_tranmiderr1'].notna() &
                               filtered_table['pl_tranmiderr2'].notna()
                               )
        dropped_targets = filtered_table[~condition_composite]
        
        # WHY IS NASA COMPOSITE TABLE INCLUDING VALUES WITHOUT ERRORBARS AS MOST PRECISE!!!!!
        filtered_table = filtered_table[condition_composite]
        
        
        if composite:
            filtered_table = filtered_table.loc[:,
                                                [
                                                'pl_name',
                                                'pl_orbper', 
                                                'pl_orbpererr1', 
                                                'pl_orbpererr2', 
                                                'pl_trandur',
                                                'pl_trandurerr1',
                                                'pl_trandurerr2',
                                                'pl_tranmid',
                                                'pl_tranmiderr1',
                                                'pl_tranmiderr2',
                                                'ra',
                                                'dec',
                                                'sy_vmag'
                                                ]
                                                ]
            filtered_table['pl_refname'] = ['Composite']*len(filtered_table)
            if len(filtered_table) != original_length:
                logger.critical('Some of the planet rows were dropped because not all necessary values were found in the table. Dropped targets:')
                for ind, row in dropped_targets.iterrows():
                    name = row['pl_name']
                    logger.critical(f'    {name}')
        else:
            filtered_table.loc[:, 'pl_refname'] = [re.sub('<[^<]+?>', '', ref_name) for ref_name in filtered_table['pl_refname'].values] 
            
        return filtered_table
    
    def _add_transit_length_if_missing(self,
                                       filtered_table_all: pd.DataFrame,
                                       filtered_table_composite:pd.DataFrame) -> pd.DataFrame:
        """
        Add transit length to rows if it is missing using the composite table.

        Parameters
        ----------
        filtered_table_all : pd.DataFrame
            Filtered NASA Exoplanet Full table.
        filtered_table_composite : pd.DataFrame
            Filtered NASA Exoplanet Composite table.

        Returns
        -------
        pd.DataFrame
            Updated table with the missing transit lengths.
        """
        
        
        condition_no_transit_length = filtered_table_all['pl_trandur'].isna()
        
        
        for ind, row in filtered_table_all.loc[condition_no_transit_length].iterrows():
            pl_name = row['pl_name']
            composite_row = filtered_table_composite[filtered_table_composite['pl_name'] == pl_name]
            T14 =  composite_row['pl_trandur'].values
            if not(np.isfinite(T14)):
                logger.warning(f'Transit length not found in composite table for planet {pl_name}')
                continue
            
            filtered_table_all[filtered_table_all['pl_name'] == pl_name].fillna(value = T14[0])
            
        
        return filtered_table_all
    
    def _calculate_transit_windows_for_single_planet(self,
                                                     row: pd.core.series.Series,
                                                     date: astime.Time
                                                     ) -> list:
        """
        Calculate transit windows for single planet.

        Parameters
        ----------
        row : pd.core.series.Series
            Row of pd.DataFrame holding all the information on the planet.
        date : astime.Time
            Semester in which we want to observe.

        Returns
        -------
        transit_windows : list
            List of all posible transit window.

        """
        P = row['pl_orbper'] * u.day
        P_error = np.max([row['pl_orbpererr1'], row['pl_orbpererr2']])
        T_C = astime.Time(row['pl_tranmid'], format = 'jd')
        T_C_error = np.max([row['pl_tranmiderr1'], row['pl_tranmiderr2']])
        transit_windows = []
        n = 0
        
        while date[0] < T_C: # To ensure we search from beginning. This will add some uncertainties if searching for past windows, but that is generally irrelevant in the purpose of this code case
            T_C -= P
            n +=1
            logger.warning('Please use T_C that is less than the starting date of observations. Otherwise, the error propagation will be overestimated.')

        while date[1] > T_C:
            T_C += P
            n +=1
            if date[0] < T_C and date[1] > T_C:
                sigma_T0 = np.sqrt(T_C_error**2 + n**2 * P_error**2)
                transit_windows.append([T_C, sigma_T0*u.day])
                
        return transit_windows
        
    
    def calculate_transit_windows(self,
                                  exoplanet_table: pd.DataFrame,
                                  date: astime.Time
                                  ):
        """
        Calculate all transit windows for all planet/references within given period.

        Parameters
        ----------
        exoplanet_table : pd.DataFrame
            Table of all references to use.
        date : astime.Time
            When to consider transit windows.

        Returns
        -------
        None.

        """
        for ind, row in exoplanet_table.iterrows():
            transit_windows = self._calculate_transit_windows_for_single_planet(
                    row = row,
                    date = date
                    )
            
            
            self.planets.append(
                _Planet_transits(
                    name = row['pl_name'],
                    transit_windows = transit_windows,
                    parameters = row
                    )
                )
        return
    
    
    def _print_info_about_window(self,
                                 Planet_transit: _Planet_transits
                                 ):
        
        logger.info('Currently working on %s'%(Planet_transit.name) + ' with reference %s'%(Planet_transit.parameters['pl_refname']))
        logger.info('P = {period:.6f} d'.format(period = Planet_transit.parameters['pl_orbper']) +
                    u" \u00B1 " + '{period_error:.6f}'.format(
                        period_error = max([Planet_transit.parameters['pl_orbpererr1'],
                                            Planet_transit.parameters['pl_orbpererr2']])
                        )
                    )
        logger.info('T14 = {t14:.3f} h'.format(t14 = Planet_transit.parameters['pl_trandur'])  +
                    u" \u00B1 " + '{t14_error:.3f}'.format(
                        t14_error = max([Planet_transit.parameters['pl_trandurerr1'],
                                            Planet_transit.parameters['pl_trandurerr2']])
                        )
                    )
        logger.info('Tc ' + '= {tc:.6f} [JD]'.format(tc = Planet_transit.parameters['pl_tranmid']) +
                    u" \u00B1 " + '{tc_error:.6f}'.format(
                        tc_error = max([Planet_transit.parameters['pl_tranmiderr1'],
                                        Planet_transit.parameters['pl_tranmiderr2']])
                        )
                    )
        logger.info('='*60)
            
        
        return

    def plot_transit_windows(self,
                             Location: str,
                             partial: bool | float = False,
                             ):
        """
        Creates plots of all valid transit windows provided Location, assuming the filtered table and all possible ephemeris sets.

        Parameters
        ----------
        Location : str
            Location where we wish to observe. Must be implemented in the observatories.Observatories class
        partial : bool
            Whether include partial transits in the planning.

        Returns
        -------
        None.

        """
        logger.info('='*60)
        
        # if partial:
        #     logger.critical('Partial windows are not implemented yet!')
        self._order_list()

        
        for Planet_transit in self.planets:
            if len(Planet_transit.transit_windows) == 0:
                logger.warning('No transit window found for planet %s'%Planet_transit.parameters['pl_name'])
                continue
            self._print_info_about_window(Planet_transit)
            best_reference = self._find_best_ephemeris(Planet_transit= Planet_transit)
            
            if best_reference == Planet_transit.parameters['pl_refname']:
                path_file_reference = Location + '/%s'%(Planet_transit.name) + '/best_%s'%(Planet_transit.parameters['pl_refname'])
                txt_summary =  Location + '/%s'%(Planet_transit.name) + '/best_%s'%(Planet_transit.parameters['pl_refname']) + '.csv'
            else:
                path_file_reference = Location + '/%s'%(Planet_transit.name) + '/%s'%(Planet_transit.parameters['pl_refname'])
                txt_summary =  Location + '/%s'%(Planet_transit.name) + '/best_%s'%(Planet_transit.parameters['pl_refname']) + '.csv'
                
            path_file_reference = path_file_reference.replace(' ', '')
            txt_summary = txt_summary.replace(' ', '')
            
            for night, uncertainty in Planet_transit.transit_windows:
                new_Plot = Plot_transit(Location = Location,
                                          Planet_transit = Planet_transit,
                                          Night = night,
                                          Uncertainty = uncertainty,
                                          partial= partial
                                          )
                
                if new_Plot.quality != -9999:
                    semester = self._check_P_ESO(new_Plot.Night)
                    file_name = self.save_directory + '/' + path_file_reference + '/Q' + str(new_Plot.quality) + '_'  + f'{semester}_'+ (new_Plot.Night-1*u.day).to_value('datetime').strftime('%Y%m%d')
                    # Make directory if it doesn't exist
                    os.makedirs(self.save_directory + '/' + path_file_reference, mode = 0o777, exist_ok = True) 
                    new_Plot.save_transit_window(file_name)
                    
                    with open(self.save_directory + '/' +txt_summary, 'a') as f:
                        line = Planet_transit.name + ';' + (new_Plot.Night-1*u.day).to_value('datetime').strftime('%Y%m%d') + ';' + str(new_Plot.quality) + ';' + new_Plot.TimeArray.time_array[new_Plot.ObservationsFull.complete[0]].to_value(
                             format= 'datetime').strftime('%H:%M') + ';' + new_Plot.TimeArray.time_array[new_Plot.ObservationsFull.complete[-1]].to_value(
                             format= 'datetime').strftime('%H:%M') + ';'+ str(int(len(new_Plot.VisibilityFull.Ingress)>30))  + '\n' 
                        f.write(line)
                    
                else:
                    night = (new_Plot.Night-1*u.day).to_value('datetime').strftime('%Y%m%d')
                    logger.warning(f'Skipping transit window {night}')
            logger.info('='*60)
        return
    
    def _find_best_ephemeris(self,
                             Planet_transit: _Planet_transits,
                             ) -> str:
        """
        Finds which of the available ephemeris is the best.

        Parameters
        ----------
        Planet_transit : _Planet_transits
            Planet transits class, including the 

        Returns
        -------
        best_reference : str
            Best reference for given transit window period
        """
        best_reference = None
        for ind, planet in enumerate(self.planets):
            if (best_reference is None and 
                Planet_transit.parameters['pl_name'] == planet.parameters['pl_name']):
                best_reference = planet.parameters['pl_refname']
                uncertainty = planet.transit_windows[0][1]
            elif Planet_transit.parameters['pl_name'] == planet.parameters['pl_name']:
                if planet.transit_windows[0][1] < uncertainty:
                    best_reference = planet.parameters['pl_refname']
                    uncertainty = planet.transit_windows[0][1]
            else:
                continue

        return best_reference
    
    def _order_list(self):
        """
        Order list of planet ephemeris by planet name. This ensures the calculation is done planet by planet.
        """
        
        order = []
        for Planet_transit in self.planets:
            order.append(Planet_transit.name)
            indices = np.argsort(np.array(order))
        new_list = []
        # CLEANME: This is dumb.
        for ind in indices:
            new_list.append(self.planets[ind])
            
        self.planets = new_list
        return
    
    
    def _check_P_ESO(self,
                     Night: astime.Time) -> str:
        """
        Calculates in which period the transit night is to order the transit based on it.,

        Parameters
        ----------
        Night : astime.Time
            Night of the transit

        Returns
        -------
        semester : str
            Which period of ESO proposal call does this window fall in.
        """
        P112 = astime.Time(['2024-04-01 12:00:00.000'],scale = 'utc')
        P113 = astime.Time(['2024-10-01 12:00:00.000'],scale = 'utc')
        semester = None
        ind = 0
        while semester is None:
            if Night > P113:
                P113 += 1*u.year
                P112 += 1*u.year
                ind += 2
            elif Night > P112:
                P112 += 1*u.year
                ind += 1
            else:
                semester = 112+ind

        return str(semester)
    
    
    def create_transit_summary(self,
                               Plots,
                               filename:str
                               ):
        information = {}
        
        for plot in Plots:
            baseline_full = ('between(' +
                             self.full_observation_time[0].strftime('%Y-%m-%dT%H:%M') + ', '+ 
                             self.full_observation_time[1].strftime('%Y-%m-%dT%H:%M') + ', '+
                             str(plot.quality) + ', "Transit of %s; '%(plot.Planet_transit.name) +
                             '{time:.2f}'.format(time = (
                                 plot.full_observation_time[1] - 
                                 plot.full_observation_time[0]).seconds/3600
                                 ) + ' hours of observation"'
                             )
            baseline_twilight = None
            if not(plot.twilight_observation_time is None):
                baseline_twilight = ('between(' +
                                 self.twilight_observation_time[0].strftime('%Y-%m-%dT%H:%M') + ', '+ 
                                 self.twilight_observation_time[1].strftime('%Y-%m-%dT%H:%M') + ', '+
                                 str(plot.quality) + ', "Transit of %s; '%(plot.Planet_transit.name) +
                                 '{time:.2f}'.format(time = (
                                     plot.twilight_observation_time[1] - 
                                     plot.twilight_observation_time[0]).seconds/3600
                                     ) + ' hours of observation"'
                                 )
            
            try:
                information[str(plot.quality)].append({
                    'Night': plot.Night.to_value('datetime').strftime('%Y%m%d'),
                    'Baseline_full': baseline_full,
                    'Baseline_twilight': baseline_twilight,
                    'filename': plot.filename,
                    'quality': plot.quality
                    })
            except:
                information[str(plot.quality)] = [{
                    'Night': plot.Night.to_value('datetime').strftime('%Y%m%d'),
                    'Baseline_full': baseline_full,
                    'Baseline_twilight': baseline_twilight,
                    'filename': plot.filename,
                    'quality': plot.quality
                    }]
            pass
        
        
        
        return
        
        
#%% Get dates for ESO semester counting
def get_dates_for_ESO_semester(
        P_number: int,
        large_program: bool = False
        ) -> astime.Time:
    """
    Provide starting and ending date of the ESO semester.

    Parameters
    ----------
    P_number : int
        Which semester are we looking for.
    large_program : bool, optional
        If we consider a large (4 semesters) program. Will give the 6 semesters by default, as ESO asks for transit windows in two more semesters. The default is False.

    Returns
    -------
    time : astime.Time
        Starting and ending date of the semester.

    """
    P112 = astime.Time(['2023-10-01 12:00:00.000', '2024-04-01 12:00:00.000'],
                       scale = 'utc',)
    P113 = astime.Time(['2024-04-01 12:00:00.000', '2024-10-01 12:00:00.000'],
                       scale = 'utc',)
    P112_large = astime.Time(['2023-10-01 12:00:00.000', '2026-10-01 12:00:00.000'],
                             scale = 'utc',)
    P113_large = astime.Time(['2024-04-01 12:00:00.000', '2027-04-01 12:00:00.000'],
                             scale = 'utc',)
    
    if P_number%2 == 0:
        if large_program:
            time = P112_large + ((P_number - 112)/2)*u.year
        else:
            time = P112 + ((P_number - 112)/2)*u.year
    else:
        if large_program:
            time = P113_large + ((P_number - 113)/2)*u.year
        else:
            time = P113 + ((P_number - 113)/2)*u.year
    return time



from dataclasses import dataclass, field
import pandas as pd
import astropy.time as astime
import astropy.units as u
import logging
import numpy as np
from ..utils.utilities import logger_default
from datetime import datetime, timedelta
from .observability import Event
from ..telescopes.telescopes import Telescope, VLT

logger = logging.getLogger(__name__)
logger = logger_default(logger) 

def define_baseline(table):
    array_length = len(table)
    table['Planet.Baseline'] = np.nanmax([np.nanmin([[3]*array_length, .75*table['Planet.TransitDuration']], axis = 0),[2]*array_length],                                            axis=0)
    return table


@dataclass
class Windows:
    table: pd.DataFrame
    center_phase: float = 0.0
    observing_period: astime.Time | None | str = None
    baseline: u.Quantity | None = None
    large_program: bool = False
    directory: str = ''
    Airmass_limit: float = None
    
    windows_in_period:list =  field(default_factory=list)
    
    
    def generate_windows(self):
        
        if len(self.table) == 0:
            raise ValueError('Table is empty. Cannot calculate windows')
        else:
            logger.info(f'About to calculate event midpoints for {len(self.table)} planets')
        
        Ti_values = []
        Ti_sigma_values = []
        
        T1, T2 = self.observing_period[0], self.observing_period[1]
        
        for _, row in self.table.iterrows():
            T_c = row['Planet.TransitMidpoint'] * u.day
            T_c_error = np.max([row['Planet.TransitMidpoint.Error.Lower'], row['Planet.TransitMidpoint.Error.Upper']]) * u.day
            P = row['Planet.Period'] * u.day
            P_error = np.max([row['Planet.Period.Error.Lower'], row['Planet.Period.Error.Upper']]) * u.day
            
            n_min = np.ceil(((T1 - T_c).jd)*u.day / P)  # Round up to next integer
            n_max = np.floor(((T2 - T_c).jd)*u.day / P)  # Round down to previous integer

            n_values = np.arange(n_min, n_max + 1, dtype=int)

            Ti = T_c + n_values * P
            sigma_values = np.sqrt(T_c_error**2 + n_values**2 * P_error**2)
            # Shouldn't change, but to be sure we are within the observing period
            valid_Ti = Ti[(Ti > T1.jd *u.day) & (Ti < T2.jd*u.day)]
            valid_sigma = sigma_values[(Ti > T1.jd*u.day) & (Ti < T2.jd*u.day)].to(u.min)

            valid_Ti = [astime.Time(t, format='jd') for t in valid_Ti]
            
            Ti_values.append(valid_Ti)
            Ti_sigma_values.append(valid_sigma)

        self.table['Planet.TransitWindowCenter'] = Ti_values
        self.table['Planet.TransitWindowCenter.Error'] = Ti_sigma_values

        return
    
    def print_windows(self):
        for _, row in self.table.iterrows():
            logger.print('='*25)
            logger.print(f"Transit windows centers for {row['Planet.Name']}")
            for window, uncertainty in zip(row['Planet.TransitWindowCenter'], row['Planet.TransitWindowCenter.Error']):
                logger.print(f"    {window} ± {uncertainty:.2f} | {window.strftime('%Y-%m-%d %H:%M')} UT")
    
    def __post_init__(self):
        self.get_proposal_period()
        logger.info('='*25)
        logger.info('Set observing period:')
        logger.info(f'    {self.observing_period[0]}')
        logger.info(f'    {self.observing_period[1]}')
        logger.info('='*25)
        
        self.generate_windows()
        self.define_baseline()
        
        
    
    
    def get_proposal_period(self):
        if self.observing_period is None:
            self._get_period_for_next_year()
        elif type(self.observing_period) == str:
            self._get_proposal_period_from_string()
        elif type(self.observing_period) == astime.Time:
            return
        else:
            raise ValueError('Invalid type for observing period.') 

    def _get_period_for_next_year(self):
        today = datetime.now().replace(hour=12, minute=0, second=0, microsecond=0)
        next_year = today.replace(year=today.year + 1)

        date1 = today.strftime('%Y-%m-%d 12:00:00.000')
        date2 = next_year.strftime('%Y-%m-%d 12:00:00.000')
        
        self.observing_period = astime.Time([date1, date2], scale='utc')

    def _get_proposal_period_from_string(self):
        if self.observing_period.startswith('ESO.'):
            self.observing_period = self.get_dates_for_ESO_semester(
                P_number= int(''.join(c for c in self.observing_period if c.isdigit())),
            )
        else: 
            raise ValueError('Invalid string format.')

    
    def get_dates_for_ESO_semester(self,
                                   P_number: int,
                                   ) -> astime.Time:
        """
        Provide starting and ending date of the ESO semester.

        Parameters
        ----------
        P_number : int
            Which semester are we looking for.
            
        Returns
        -------
        time : astime.Time
            Starting and ending date of the semester.

        """
        if not(self.large_program):
            if P_number%2 ==0:
                time = astime.Time([f'{1967+ int(P_number/2)}-10-01 12:00:00.000',  
                                    f'{1968+int(P_number/2)}-04-01 12:00:00.000'], scale = 'utc')
            else:
                time = astime.Time([f'{1968 + int((P_number-1)/2)}-04-01 12:00:00.000',
                                    f'{1968+ int((P_number-1)/2)}-10-01 12:00:00.000'], scale = 'utc')
        else:
            if P_number%2 ==0:
                time = astime.Time([f'{1967+ int(P_number/2)}-10-01 12:00:00.000',  
                                    f'{1969+ int(P_number/2)}-10-01 12:00:00.000'], scale = 'utc')
            else:
                time = astime.Time([f'{1968 + int((P_number-1)/2)}-04-01 12:00:00.000',
                                    f'{1970 + int((P_number-1)/2)}-04-01 12:00:00.000'], scale = 'utc')
        return time
    
    
    def generate_observability(self,
                               location: Telescope,
                               partial: float = 1):
        
        if not(self.Airmass_limit):
            if location.name == 'Very Large Telescope (VLT)':
                logger.info('Setting airmass limit to 2.2.')
                self.Airmass_limit = 2.2
            else:
                logger.warning('No airmass limit set. Setting to 2')
                self.Airmass_limit = 2
        
        for _, row in self.table.iterrows():
            logger.print('='*25)
            logger.print(f"Working on {row['Planet.Name']}")
            logger.print(f"    with Tc: {row['Planet.TransitMidpoint']}, P: {row['Planet.Period']} days and T14: {row['Planet.TransitDuration']} hours")
            logger.print('='*25)
            
            
            for window, window_uncertainty in zip(row['Planet.TransitWindowCenter'], row['Planet.TransitWindowCenter.Error']):
                self.windows_in_period.append(
                    Event(
                        Location=location,
                        TransitDuration= row['Planet.TransitDuration'],
                        baseline=row['Planet.Baseline'],
                        Night= window,
                        row= row,
                        directory= self.directory,
                        Uncertainty= window_uncertainty,
                        partial= partial,
                        Airmass_limit= self.Airmass_limit,
                    )
                )

    def define_baseline(self):
        
        if self.baseline is None:
            self.table = define_baseline(self.table)
        else:
            self.table['Planet.Baseline'] = self.baseline.to(u.hour)
        
        return



if __name__ == '__main__':
    import os
    from ..database.NASA_exoplanet_archive import NASA_Exoplanet_Archive_CompositeDefault
    
    os.chdir('/media/chamaeleontis/Observatory_main/Code/observations_transits/PTO/')
    test = NASA_Exoplanet_Archive_CompositeDefault()
    test.load_API_table(force_load=True)
    
    logger.print(f"Length before further filtering of the table: {test.table.shape[0]}")
    test.table = test.table[test.table['Magnitude.V'] < 10]
    test.table = test.table[test.table['Planet.RadiusEarth'] > 3]
    test.table = test.table[test.table['Planet.RadiusEarth'] < 8]
    test.table = test.table[test.table['Planet.Period'] < 30]
    logger.print(f"Length after further filtering of the table: {test.table.shape[0]}")

    
    Transits = Windows(
        table = test.table,
        observing_period = 'ESO.115',
        directory= '/media/chamaeleontis/Observatory_main/ESO_scheduling/PTO_developement/',
        large_program= False
    )
    
    Transits.print_windows()
    
    Transits.generate_observability(
        location= VLT
    )
    
    logger.print('General Kenobi!!!!')


# #!/usr/bin/env python3
# # -*- coding: utf-8 -*-
# """
# Created on Fri Aug 18 13:45:10 2023

# @author: chamaeleontis
# """

# #%% Importing libraries
# import datetime
# import pandas as pd
# import pyvo as vo
# import astropy
# import astropy.constants as con
# import astropy.units as u
# import matplotlib
# import matplotlib.pyplot as plt
# from dataclasses import dataclass, asdict, field
# import pickle
# import logging
# import numpy as np
# import os
# import seaborn as sns
# from colorlog import ColoredFormatter
# from PTO.utilities import logger_default
# logger = logging.getLogger(__name__)
# logger = logger_default(logger) 

# #%% Convert_Earth_to_Jupiter
# def rad_earth2jupiter(x): # Convert radius of Earth to Jupiter
#     return x * con.R_earth.to(u.R_jup).value

# def rad_jupiter2earth(x): # Convert radius Jupiter to Earth
#     return x * con.R_jup.to(u.R_earth).value

# def mass_earth2jupiter(x): # Convert mass Earth to Jupiter
#     return x * con.M_earth.to(u.M_jup).value

# def mass_jupiter2earth(x): # Convert mass Jupiter to Earth
#     return x * con.M_jup.to(u.M_earth).value

# #%% NASA_exoplanets_archive
# class NASA_Exoplanets_archive:
#     """Class holding information extracted from the NASA Composite table."""
    
#     def _save(self):
#         """
#         Save class as a pickle file to predetermined location.

#         Returns
#         -------
#         None.

#         """
#         os.makedirs('./saved_files/', mode = 0o777, exist_ok = True) 
#         with open('./saved_files/NASA_exoplanets_archive.pkl', 'wb') as output_file:
#             pickle.dump(self.__dict__, output_file)


#     def _load(self):
#         """
#         Load class from pickle file in predetermined location.

#         Returns
#         -------
#         None.

#         """
#         with open('./saved_files/NASA_exoplanets_archive.pkl', 'rb') as input_file:
#             self.__dict__  =  pickle.load(input_file)
    
#     def _load_API_table(self, full = False):
#         """
#         Load the Planetary Systems (PS) and Planetary System Composite (PSComp) tables.

#         Returns
#         -------
#         None.

#         """
#         service = vo.dal.TAPService("https://exoplanetarchive.ipac.caltech.edu/TAP/") # Initialization of TAP service
#         # Searching relevant parameters
#         self.nasa_table_composite = pd.DataFrame(service.search("SELECT * FROM pscomppars"))
#         if full:
#             self.nasa_table_all = pd.DataFrame(service.search("SELECT * FROM ps"))
#         else:
#             self.nasa_table_all = pd.DataFrame(service.search("SELECT pl_name, pl_tranmid,pl_tranmiderr1, pl_tranmiderr2, pl_orbper, pl_orbpererr1, pl_orbpererr2, pl_trandur, pl_trandurerr1, pl_trandurerr2, pl_refname, dec, sy_vmag, ra FROM ps"))
#         self.time = datetime.datetime.now()
#         self._get_most_precise()
#         self._add_Earth_and_Jupiter_units()
#         self._calculate_insolation_flux()
#         self._calculate_scale_height()
#         self._calculate_transit_length()
       
        
        
#         self.filtered_table_composite = self.nasa_table_composite
#         self.filtered_table_all = self.nasa_table_all
#         self._save()
        
    
    
#     def __init__(self,
#                  force_reload: bool = False,
#                  use_PS: bool = True,
#                  full_PS: bool = False):
#         """
#         Initialize the class. This will try to load the NASA table already saved in predetermined location. If it fails, it will extract the table from NASA archive through TAP service. In case the table is loaded, but the loading was done more than week ago, the table will be reloaded automatically.
    
#         Parameters
#         ----------
#         force_reload : bool, optional
#             If turned to True it forces reload of the table. The default is False.
    
#         Returns
#         -------
#         None.
    
#         """
#         try: # Test loading the function
#             logger.info('Trying to load NASA table')
#             self._load()
#             if (datetime.datetime.now() - self.time).days > 7 or force_reload: # If not < week data or requested, reload anyway
#                 logger.info('Reloading NASA table [new week] or forced reload. This can take a long time if using the full (PS) table (around 10 min)')
#                 self._load_API_table(full_PS)
#         except: # Loading failed, reload the archive - should happen only the first time running the code
#             logger.warning('Weird, no table found. This should happen only for the first time running the code. This also takes a long time if loading both Composite (PSComp) and Full (PS) tables (around 10 min)')
#             self._load_API_table(full_PS)
        
        
        
#         if not(use_PS): # Not the most effective way of solving
#             self.nasa_table_all = self.nasa_table_all[0:0]
#             self.filtered_table_all = self.filtered_table_all[0:0]
#         return

    
#     def _add_Earth_and_Jupiter_units(self):
#         """
#         Add planetary radii in units of Earth/ Jupiter if only one version exists.

#         Returns
#         -------
#         None.

#         """
# # =============================================================================
# #         Earth to Jupiter radius
# # =============================================================================
#         condition_composite = (self.nasa_table_composite['pl_rade'].notna() & # Planetary radius [Earth units]
#                                self.nasa_table_composite['pl_radj'].isna() # Planetary radius [Jupiter units]
#                                 )
#         indices_composite = self.nasa_table_composite[condition_composite].index
        
#         self.nasa_table_composite.loc[indices_composite, 'pl_radj'] = (
#             self.nasa_table_composite.loc[indices_composite, 'pl_rade'] * con.R_jup / con.R_earth
#             )
        
# # =============================================================================
# #         Jupiter to Earth radius
# # =============================================================================
#         condition_composite = (self.nasa_table_composite['pl_radj'].notna() & # Planetary radius [Jupiter units]
#                                self.nasa_table_composite['pl_rade'].isna() # Planetary radius [Earth units]
#                                 )
#         indices_composite = self.nasa_table_composite[condition_composite].index
        
#         self.nasa_table_composite.loc[indices_composite, 'pl_rade'] = (
#             self.nasa_table_composite.loc[indices_composite, 'pl_radj'] * con.R_earth / con.R_jup
#             )
#         return

# # =============================================================================
# #     Calculation of additional parameters
# # =============================================================================
#     def _calculate_transit_length(self):
#         r"""
#         Calculates the transit length.
        
#         Equations used:
#             T_{14} = \frac{P}{\pi} * \arcsin(\frac{\frac{R_s}{a} * \sqrt{(1 + \frac{R_p}{R_s})^2 - b^2)}}{\sin(i)})
#             T_{14} = \frac{P}{\pi} * \arcsin{(\frac{\frac{R_s}{a} * \sqrt{1+ (\frac{R_p}{R_s})^2} - b^2}{\sin{i}})}
        
#         In case of eccentricity, correction factor is used:
#             \frac{\sqrt{1-e^2}}{(1+e)\sin(\omega)} 
#             \frac{\sqrt{1-e^2}}{(1+e)} 

#         Returns
#         -------
#         None.

#         """
#         logger.critical('Implement uncertainties when calculating transit length')
#         condition_composite = (
#             self.nasa_table_composite['pl_trandur'].isna() & # Transit duration
#             self.nasa_table_composite['pl_orbper'].notna() & # Orbital period
#             self.nasa_table_composite['pl_imppar'].notna() & # Impact parameter
#             self.nasa_table_composite['pl_orbincl'].notna() & # Orbital inclination
#             self.nasa_table_composite['pl_ratdor'].notna() & # Ratio of semimajor axis to stellar radius (a/Rs)
#             self.nasa_table_composite['pl_ratror'].notna() # Ratio of planet to stellar ratio (Rp/Rs)
#             )
        
#         indices_composite = self.nasa_table_composite[condition_composite].index
#         indices_composite_combined = indices_composite
        
# # =============================================================================
# #         $$
# #         T_{14} = \frac{P}{\pi} * \arcsin(\frac{\frac{R_s}{a} * \sqrt{(1 + \frac{R_p}{R_s})^2 - b^2)}}{\sin(i)})
# #         $$
# # =============================================================================
#         self.nasa_table_composite.loc[indices_composite, 'pl_trandur'] = (
#             (
#             (self.nasa_table_composite.loc[indices_composite, 'pl_orbper'] * 24 / np.pi)*
#             np.arcsin(
#                     (1./
#                      self.nasa_table_composite.loc[indices_composite, 'pl_ratdor']
#                      ) *
#                     np.sqrt((1. +
#                              self.nasa_table_composite.loc[indices_composite, 'pl_ratror'])**2. -
#                             self.nasa_table_composite.loc[indices_composite, 'pl_imppar']**2.
#                             ) /
#                     np.sin((self.nasa_table_composite.loc[indices_composite, 'pl_orbincl']*np.pi/180.))
#                     )
#             )
#             )
        
#         condition_composite = (
#             self.nasa_table_composite['pl_trandur'].isna() & # Transit duration
#             self.nasa_table_composite['pl_orbper'].notna() & # Orbital period
#             self.nasa_table_composite['pl_imppar'].notna() & # Impact parameter
#             (self.nasa_table_composite['pl_imppar'] >= 0) & # Impact parameter
#             (self.nasa_table_composite['pl_imppar'] <= 1) & # Impact parameter
#             self.nasa_table_composite['pl_orbincl'].notna() & # Orbital inclination
#             self.nasa_table_composite['pl_orbsmax'].notna() & # Semi-major axis
            
#             self.nasa_table_composite['pl_rade'].notna() & # Planetary radius [Earth units]
#             self.nasa_table_composite['st_rad'].notna() # Stellar radius [Solar units]
#             )
        
        
#         indices_composite = self.nasa_table_composite[condition_composite].index

        
# # =============================================================================
# #         $$
# #         T_{14} = \frac{P}{\pi} * \arcsin{(\frac{\frac{R_s}{a} * \sqrt{1+ (\frac{R_p}{R_s})^2} - b^2}{\sin{i}})}
# #         $$
# # =============================================================================
#         self.nasa_table_composite.loc[indices_composite, 'pl_trandur'] = (
#             (self.nasa_table_composite.loc[indices_composite, 'pl_orbper'] * 24 / np.pi) *
#             np.arcsin(
#                         (
#                     self.nasa_table_composite.loc[indices_composite, 'st_rad']  /
#                     self.nasa_table_composite.loc[indices_composite, 'pl_orbsmax'] *
#                     con.R_sun / con.au 
#                             ) *
#                     np.sqrt((1. +
#                             (
#                             (self.nasa_table_composite.loc[indices_composite, 'pl_rade'])/
#                             self.nasa_table_composite.loc[indices_composite, 'st_rad'] *
#                             con.R_earth / con.R_sun)**2 -
#                             self.nasa_table_composite.loc[indices_composite, 'pl_imppar']**2.
#                             ) /
#                     np.sin((self.nasa_table_composite.loc[indices_composite, 'pl_orbincl']*np.pi/180.))
#                             )
#                     )
#             )
        
#         indices_composite_combined = indices_composite_combined.union(indices_composite)
# # =============================================================================
# #         Eccentricity correction
# # =============================================================================
#         condition_composite = (
#             (self.nasa_table_composite['pl_orbeccen'] > 0.) & # Eccentricity
#             self.nasa_table_composite['pl_orblper'].notna() # Argument of periastron
#             )
        
#         indices_composite = self.nasa_table_composite[condition_composite].index
#         indices_composite = indices_composite_combined.intersection(indices_composite)
        
# # =============================================================================
# #         $$
# #         T_{14} = T_{14} * \frac{\sqrt{1-e^2}}{(1+e)\sin(\omega)} 
# #         $$
# # =============================================================================
#         self.nasa_table_composite.loc[indices_composite, 'pl_trandur'] *= np.sqrt(
#             1. - self.nasa_table_composite.loc[indices_composite,'pl_orbeccen']**2.
#             ) / (
#                 1. + (self.nasa_table_composite.loc[indices_composite, 'pl_orbeccen'] * 
#                 np.sin(self.nasa_table_composite.loc[indices_composite, 'pl_orblper'] * 
#                        np.pi / 180.))
#                 )
        
        
#         condition_composite = (
#             (self.nasa_table_composite['pl_orbeccen'] > 0.) & # Eccentricity
#             self.nasa_table_composite['pl_orblper'].isna() # Argument of periastron
#             )
        
#         indices_composite = self.nasa_table_composite[condition_composite].index
#         indices_composite = indices_composite_combined.intersection(indices_composite)
        
# # =============================================================================
# #         $$
# #         T_{14} = T_{14} * \frac{\sqrt{1-e^2}}{(1+e)} 
# #         $$
# # =============================================================================
#         self.nasa_table_composite.loc[indices_composite, 'pl_trandur'] *= np.sqrt(
#             1. - self.nasa_table_composite['pl_orbeccen']**2.
#             ) / (1. + self.nasa_table_composite.loc[indices_composite, 'pl_orbeccen'])
        
#         return
    
    
#     def _calculate_insolation_flux(self):
#         """
#         Calculate insolation flux for planets with sufficient parameters.

#         Returns
#         -------
#         None.

#         """
#         condition_composite = (self.nasa_table_composite['pl_insol'].isna() & 
#                                self.nasa_table_composite['st_lum'].notna() &
#                                self.nasa_table_composite['pl_orbsmax'].notna()
#                                )
        
#         indices_composite = self.nasa_table_composite[condition_composite].index
        
#         self.nasa_table_composite.loc[indices_composite, 'pl_insol'] = (
#             10**self.nasa_table_composite.loc[indices_composite, 'st_lum'] / 
#             self.nasa_table_composite.loc[indices_composite,'pl_orbsmax']**2
#             )
        
#         return
    
#     def _calculate_scale_height(self,
#                                molecular_mass:float = 2.3):
#         """
#         Calculate atmospheric scale heights for all planets with sufficient parameters, assuming molecular mass of 2.3kg/mol as default.

#         Parameters
#         ----------
#         mol_mass : float, optional
#             Molecular mass in kilograms per mol [kg/mol]. The default is 2.3.

#         Returns
#         -------
#         None.

#         """
#         try:
#             self.nasa_table_composite['H']
#         except:
#             self.nasa_table_composite['H'] = [np.nan]*len(self.nasa_table_composite)
        
#         condition_composite = (
#                             self.nasa_table_composite['H'].isna() &
#                             self.nasa_table_composite['pl_bmassj'].notna() &
#                             self.nasa_table_composite['pl_radj'].notna() &
#                             self.nasa_table_composite['pl_eqt'].notna()
#                             )
        
#         indices_composite = self.nasa_table_composite[condition_composite].index
        
#         logger.critical('Atmospheric scale height calculation is not debugged yet!')
#         self.nasa_table_composite.loc[indices_composite, 'H']  = (
#             con.N_A *
#             self.nasa_table_composite.loc[indices_composite, 'pl_eqt'] * u.K  *
#             ((self.nasa_table_composite.loc[indices_composite, 'pl_radj'] * (1*u.R_jup).to(u.m))**2) *
#             con.k_B /
#             (
#             (molecular_mass*u.kg/u.mol) *
#             con.G *
#             self.nasa_table_composite.loc[indices_composite, 'pl_bmassj'] * (1*u.M_jup).to(u.kg))
#             )
        
#         return
    
    
#     def reset_filter(self):
#         """
#         Reset the filtered_table atributes to full table.

#         Returns
#         -------
#         None.

#         """
#         self.filtered_table_composite = self.nasa_table_composite
#         self.filtered_table_all = self.nasa_table_all
#         return
    
#     def _bind_filter_all(self):
#         """
#         Bind the filter used by composite table to the all table.

#         Returns
#         -------
#         None.

#         """
#         condition = self.filtered_table_all['pl_name'].isin(
#             self.filtered_table_composite['pl_name'].values
#             )
#         self.filtered_table_all = self.filtered_table_all[condition]
#         return
    
#     def _bind_filter_composite(self):
#         """
#         Bind the filter used by composite table to the all table.

#         Returns
#         -------
#         None.

#         """
#         condition = self.filtered_table_composite['pl_name'].isin(
#             self.filtered_table_all['pl_name'].values
#             )
#         self.filtered_table_composite = self.filtered_table_composite[condition]
#         return
    
#     def filter_table_composite(self,
#                                condition: pd.core.series.Series) -> pd.DataFrame:
#         """
#         Filter the composite table by pre-defined condition indices.

#         Parameters
#         ----------
#         condition : pd.core.series.Series
#             Series with boolean values that filter the composite table.

#         Returns
#         -------
#         filtered_table_composite : pd.DataFrame
#             Filtered table.

#         """
#         self.filtered_table_composite = self.nasa_table_composite[condition]
#         self._bind_filter_all()
#         return self.filtered_table_composite
    
#     def filter_table_all(self, condition):
#         """
#         Filter the complete table by pre-defined condition indices.

#         Parameters
#         ----------
#         condition : pd.core.series.Series
#             Series with boolean values that filter the composite table.

#         Returns
#         -------
#         filtered_table_all : pd.DataFrame
#             Filtered table.

#         """
#         self.filtered_table_all = self.nasa_table_all[condition]
#         self._bind_filter_composite()
#         return self.filtered_table_all
    
#     def _check_if_fig_and_ax_exist(self,
#                                    fig: matplotlib.figure.Figure | None,
#                                    ax: plt.Axes | None) -> tuple[matplotlib.figure.Figure,
#                                                                  plt.Axes]:
#         """
#         Checks whether figure and artist is created, and will create new one if not.

#         Parameters
#         ----------
#         fig : matplotlib.figure.Figure | None
#             Variable of the figure. If None, the figure is created with single artist
#         ax : plt.Axes | None
#             Variable of the artist. If None, a single artist is created

#         Returns
#         -------
#         fig : matplotlib.figure.Figure
#             Resulting figure, which is unchanged if passed, and created if None is passed instead.
#         ax : plt.Axes
#             Resulting artist, which is unchanged if passed, and created if None is passed instead.

#         """
        
#         if fig is None:
#             fig,ax = plt.subplots(1)
#             return fig, ax 
#         else:
#             return fig, ax
    
#     def _default_plot_insolation_flux_diagram_settings(self,
#                                                        ax):
#         """Set the ax scales to log scale."""
#         ax.set_xscale('log')
#         ax.set_yscale('log')
#         return
        
    
#     def plot_radius_insolation_flux_diagram(self,
#                                             fig: matplotlib.figure.Figure | None = None,
#                                             ax: plt.Axes | None =None,
#                                             ) -> tuple[matplotlib.figure.Figure, plt.Axes]:
#         logger.critical("TODO: Rewrite this function to be more general")
#         fig, ax = self._check_if_fig_and_ax_exist(fig,ax)
        
#         self._default_plot_insolation_flux_diagram_settings(ax)
        
#         condition_composite = (self.nasa_table_composite['pl_insol'].notna() &
#                                self.nasa_table_composite['pl_rade'].notna() &
#                                (self.nasa_table_composite['pl_insol'] != 0) &
#                                (self.nasa_table_composite['pl_rade'] != 0)
#                                )
        
#         indices_composite = self.nasa_table_composite[condition_composite].index
        
#         ax.invert_xaxis()
        

#         sns.kdeplot(x = self.nasa_table_composite.loc[indices_composite, 'pl_insol'],
#                     y = self.nasa_table_composite.loc[indices_composite, 'pl_rade'],
#                     # cmap=cmap,
#                     fill=True,
#                     label='_nolegend_',
#                     bw_adjust=0.6,
#                     log_scale =[True, True],
#                     levels = [0,.0001,.001,.01,.1,.3,.5,.8,1],
#                     ax=ax
#                     )

#         ax.scatter(
#                 self.nasa_table_composite.loc[indices_composite, 'pl_insol'],
#                 self.nasa_table_composite.loc[indices_composite, 'pl_rade'],
#                 # color = 'black',
#                 # edgecolors = 'black',
#                 label='_nolegend_',
#                 alpha = 0.5,
#                 s = 20,
#             )
        
        
#         ax2 = ax.secondary_yaxis('right', functions=(rad_earth2jupiter, rad_jupiter2earth))
#         ax.set_title('Radius vs Insolation flux for known exoplanets')

#         ax.set_xlabel('Insolation flux [$S_{\oplus}$]')
#         ax2.set_ylabel('Planetary radius $[R_{Jup}]$')
#         ax.set_ylabel('Planetary radius $[R_{\oplus}]$')
        
#         ax.set_xlim(10E4,10E0)
#         ax.set_ylim(10E-1,25E0)
#         ax.get_figure().gca().set_title("")

#         # if ykey == 'pl_bmasse':
#         #     ax2 = ax.secondary_yaxis('right', functions=(mass_earth2jupiter, mass_jupiter2earth))
#         #     ax2.set_ylabel('Planetary mass $[M_{Jup}]$', fontsize = 36)
#         return fig,ax
        
    
#     def plot_mass_insolation_flux_diagram(self,
#                                           fig=None,
#                                           ax=None):
#         fig, ax = self._check_if_fig_and_ax_exist(fig,ax)
        
#         return
    
#     def create_table_observation(self):
#         return
    
#     def add_custom_ephemeris(self,
#                              Transit_midpoint: float,
#                              Transit_midpoint_lower_error: float,
#                              Transit_midpoint_upper_error: float,
#                              Period: float,
#                              Period_lower_error: float,
#                              Period_upper_error: float,
#                              Transit_length: float,
#                              Transit_length_lower_error: float,
#                              Transit_length_upper_error: float,
#                              Right_Ascension: float,
#                              Declination: float,
#                              V_magnitude: float,
#                              Reference: str,
#                              Planet_name: str
#                              ):
#         """
#         Adds a custom ephemeris to the table.

#         Parameters
#         ----------
#         Transit_midpoint : float
#             Transit midpoint of the system.
#         Transit_midpoint_lower_error : float
#             Transit midpoint minus uncertainty.
#         Transit_midpoint_upper_error : float
#             Transit midpoint plus uncertainty.
#         Period : float
#             Period of the system.
#         Period_lower_error : float
#             Period minus uncertainty.
#         Period_upper_error : float
#             Period plus uncertainty.
#         Transit_length : float
#             Transit length of the system.
#         Transit_length_lower_error : float
#             Transit length minus uncertainty.
#         Transit_length_upper_error : float
#             Transit length plus uncertainty.
#         Right_Ascension : float
#             Right ascension of the system.
#         Declination : float
#             Declination of the system.
#         V_magnitude : float
#             V magnitude of the system.
#         Reference : str
#             Reference to use for the ephemeris.
#         Planet_name : str
#             Planet name to use for the ephemeris.

#         Returns
#         -------
#         None.

#         """
        
#         parameters = {key: '' for key in self.nasa_table_all.columns}
        
#         parameters.update({
#             'pl_name': Planet_name,
#             'pl_orbper': Period, 
#             'pl_orbpererr1': Period_upper_error, 
#             'pl_orbpererr2': Period_lower_error, 
#             'pl_trandur': Transit_length,
#             'pl_trandurerr1': Transit_length_upper_error,
#             'pl_trandurerr2': Transit_length_lower_error,
#             'pl_tranmid': Transit_midpoint,
#             'pl_tranmiderr1': Transit_midpoint_upper_error,
#             'pl_tranmiderr2': Transit_midpoint_lower_error,
#             'ra': Right_Ascension,
#             'dec': Declination,
#             'sy_vmag': V_magnitude,
#             'pl_refname': Reference,
#             })
        
#         pieces = [self.nasa_table_all, pd.DataFrame(parameters,  index=[0])]
        
#         self.nasa_table_all = pd.concat(pieces)
#         try:
#             self.nasa_table_all = self.nasa_table_all.reset_index()
#         except:
#             pass
        
#         pieces = [self.filtered_table_all, pd.DataFrame(parameters,  index=[0])]
#         self.filtered_table_all = pd.concat(pieces)
#         try:
#             self.filtered_table_all = self.filtered_table_all.reset_index()
#         except:
#             pass
        
#         return

#     def _get_most_precise_row(self,
#                               planet_data: pd.DataFrame,
#                               composite_data: pd.DataFrame):
#         for key in [
#             'pl_orbper',
#             'pl_tranmid',
#             'pl_trandur'
#             ]:
#             condition = (composite_data[key].isna() |
#                          composite_data[key+'err1'].isna() |
#                          composite_data[key+'err2'].isna())
            
#             indices_composite = composite_data[condition].index
            
#             average_error = (planet_data[key+'err1'] + abs(planet_data[key+'err2']))/2
#             if average_error.idxmin() is np.nan:
#                 continue
#             composite_data.loc[indices_composite, key] = planet_data.loc[average_error.idxmin()][key]
#             composite_data.loc[indices_composite, key+'err2'] = planet_data.loc[average_error.idxmin()][key+'err1']
#             composite_data.loc[indices_composite, key+'err2'] = planet_data.loc[average_error.idxmin()][key+'err2']
        
        
#         return composite_data
    
    
#     def _get_most_precise(self):
        
#         condition = (
#             (self.nasa_table_composite['pl_tranmid'].isna() |
#             self.nasa_table_composite['pl_tranmiderr1'].isna() |
#             self.nasa_table_composite['pl_tranmiderr2'].isna() |
#             self.nasa_table_composite['pl_orbper'].isna() |
#             self.nasa_table_composite['pl_orbpererr1'].isna() |
#             self.nasa_table_composite['pl_orbpererr2'].isna() |
#             self.nasa_table_composite['pl_trandur'].isna() |
#             self.nasa_table_composite['pl_trandurerr1'].isna() |
#             self.nasa_table_composite['pl_trandurerr2'].isna()) &
#             (self.nasa_table_composite['tran_flag'] == 1) # Transiting system
#             )
#         indices = self.nasa_table_composite[condition].index
        
#         for value in self.nasa_table_composite.loc[indices, 'pl_name']:
#             updated_row = self._get_most_precise_row(planet_data= self.nasa_table_all[self.nasa_table_all['pl_name'] == value],
#                                                      composite_data= self.nasa_table_composite[self.nasa_table_composite['pl_name'] == value])
            
#             condition = (self.nasa_table_composite['pl_name'] == value)

#             indices_composite = self.nasa_table_composite[condition].index
#             self.nasa_table_composite.loc[indices_composite] = updated_row

#             # self.nasa_table_composite[self.nasa_table_composite['pl_name'] == value] = updated_row
        
#         return

# def load_csv_ephemeris(Exoplanets: NASA_Exoplanets_archive,
#                        table: str,
#                        reference_name: str):
#     logger.info(f'Loading ephemeris from table located: {table}')
    
#     read_table = pd.read_csv(table, sep= ';')
    
#     for ind, row in read_table.iterrows():
        
#         Exoplanets.add_custom_ephemeris(
#             Transit_midpoint = float(row['T0']),
#             Transit_midpoint_lower_error = float(row['T0_err']),
#             Transit_midpoint_upper_error = float(row['T0_err']),
#             Period = float(row['P']),
#             Period_lower_error = float(row['P_err']),
#             Period_upper_error = float(row['P_err']),
#             Transit_length = float(row['T14']),
#             Transit_length_lower_error = 0,
#             Transit_length_upper_error = 0,
#             Right_Ascension = float(row['ra']),
#             Declination = float(row['dec']),
#             V_magnitude = float(row['mag_V']),
#             Reference = reference_name,
#             Planet_name = row['pl_name'],
#             )
#     return





import matplotlib.pyplot as plt
import seaborn as sns
from dataclasses import dataclass
import astropy.time as astime
import datetime
import pandas as pd
from ..telescopes.telescopes import Telescope
import astropy.units as u
import numpy as np
import matplotlib.dates as md


@dataclass
class WindowPlot():
    def generate_plots(self):
        with plt.ioff():
            fig, ax = plt.subplots(1, figsize=(18, 12))
            self._plot_basic_setting(fig, ax)
            self._plot_twilight(ax)
            self._plot_Moon(ax)
            self._plot_target(ax)
            self._plot_observation(ax)
            self._write_system_parameters(ax)
            
            return fig, ax


    def _plot_basic_setting(self,
                            fig: plt.Figure,
                            ax: plt.Axes):
        """
        Plot basic settings regardless of what planet is plotted.

        Parameters
        ----------
        fig : plt.Figure
            Figure on which to plot on.
        ax : plt.Axes
            Artist on which to plot on.

        Returns
        -------
        None.

        """
        Night = (self.TimeArray.midnight-1*u.day).to_value('datetime')

        ax.set_title(
            f"Transit of {self.row['Planet.Name']} on night: {Night.strftime('%Y%m%d')} at {self.Location.name} with precision on $T_c$ of: {self.Uncertainty.to(u.min):.2f}; Quality: {self.quality}", fontsize=14)

        ax.axhline(
            np.arcsin(1/self.Airmass_limit) * 180/np.pi,
            ls='--', lw=2,
            color='darkred'
        )

        ax.grid(color='black', linestyle='dashed', linewidth=0.3, alpha=1.0)
        xfmt = md.DateFormatter('%d/%m  %H')
        ax.xaxis.set_major_formatter(xfmt)
        fig.autofmt_xdate()
        hours = md.HourLocator()
        ax.xaxis.set_major_locator(hours)

        ax.set_ylim(0, 90)
        ax2 = ax.twinx()
        ax.set_ylabel('Altitude [deg]', color='k')
        ax2.set_ylabel('Airmass', color='k')
        ax2.tick_params('y', colors='k')
        ax2.set_ylim([0, 90])
        ax2.set_yticklabels(['', 5.76, 2.92, 2.00, 1.56,
                            1.31, 1.15, 1.06, 1.02, 1.00])

        ax.set_xlim(self.TimeArray.Sunset,
                    self.TimeArray.Sunrise)

    def _plot_twilight(self,
                       ax: plt.Axes):
        """
        Overplot the twilight area.

        Parameters
        ----------
        ax : plt.Axes
            Artist on which to plot on.

        Returns
        -------
        None.

        """
        ax.fill_between(self.TimeArray.time_array.to_value(format='datetime'),
                        0, 90,
                        self.AltitudeArray.Sun.alt.value < 0,
                        color='blue',
                        alpha=0.3
                        )
        ax.fill_between(self.TimeArray.time_array.to_value(format='datetime'),
                        0, 90,
                        self.AltitudeArray.Sun.alt.value < -18,
                        color='blue',
                        alpha=0.5
                        )
        return

    def _plot_Moon(self,
                   ax: plt.Axes):
        """
        Overplot the Moon position.

        Parameters
        ----------
        ax : plt.Axes
            Artist on which to plot on.

        Returns
        -------
        None.

        """
        ax.plot(self.TimeArray.time_array.to_value(format='datetime'),
                self.AltitudeArray.Moon.alt.value,
                ls='-',
                label='Moon',
                linewidth=2,
                color='yellow',
                alpha=0.7
                )
        return

    def _plot_target(self,
                     ax: plt.Axes):
        """
        Plot the target, highlighting the transit area.

        Parameters
        ----------
        ax : plt.Axes
            Artist on which to plot.

        Returns
        -------
        None.

        """
        ax.plot(self.TimeArray.time_array.to_value(format='datetime'),
                self.AltitudeArray.target.alt.value,
                ls='-',
                label='Target',
                linewidth=4,
                color='red',
                )
        ax.plot(self.TimeArray.time_array[self.Indices.Transit].to_value(format='datetime'),
                self.AltitudeArray.target[self.Indices.Transit].alt.value,
                ls='-',
                label='Target',
                linewidth=10,
                color='red',
                )

        ax.scatter(
            self.TimeArray.time_array[self.Indices.Transit[0][0]].to_value(
                format='datetime'),
            self.AltitudeArray.target[self.Indices.Transit[0][0]].alt.value,
            color='white',
            zorder=999,
            s=200,
        )
        ax.scatter(
            self.TimeArray.time_array[self.Indices.Transit[-1]
                                      [-1]].to_value(format='datetime'),
            self.AltitudeArray.target[self.Indices.Transit[-1][-1]].alt.value,
            color='white',
            zorder=999,
            s=200,
        )

        return

    def _plot_observation(self,
                          ax: plt.Axes):

        # CLEANME
        self.twilight_observation_time = [
                    self.TimeArray.time_array[self.ObservationsTwilight.complete[0]].to_value(
                        format='datetime'),
                    self.TimeArray.time_array[self.ObservationsTwilight.complete[-1]].to_value(
                        format='datetime')
                ]
                
        if len(self.ObservationsFull.complete) != 0:
            ax.fill_between(self.TimeArray.time_array[self.ObservationsFull.complete].to_value(format='datetime'),
                0, 10,
                color='lime',
                alpha=1,
            )

            ax.text((self.TimeArray.time_array[self.ObservationsFull.complete[0]]+30*u.min).to_value(format='datetime'),
                    5,
                    'Observations (no twilight): ' +
                    self.TimeArray.time_array[self.ObservationsFull.complete[0]].to_value(
                        format='datetime').strftime('%H:%M') + ' - ' +
                    self.TimeArray.time_array[self.ObservationsFull.complete[-1]].to_value(
                        format='datetime').strftime('%H:%M'),
                    fontsize=10,
                    )
            
            self.full_observation_time = [
                self.TimeArray.time_array[self.ObservationsFull.complete[0]].to_value(
                    format='datetime'),
                self.TimeArray.time_array[self.ObservationsFull.complete[-1]
                                          ].to_value(format='datetime')
            ]
            self.twilight_full_is_same = (self.full_observation_time == self.twilight_observation_time)
        else:
            self.twilight_full_is_same = False
            
        
        if (((self.FlagsFull.baseline_ingress) or  # Ingress baseline missing
            (self.FlagsFull.baseline_egress) or  # Egress baseline missing
                (not (self.FlagsWindow.visible) and
                 self.FlagsWindow.visible_twilight)) and # Transit into twilight
            not(self.twilight_full_is_same)): # Twilight == Full observations
            
            ax.fill_between(self.TimeArray.time_array[self.ObservationsTwilight.complete].to_value(format='datetime'),
                10, 20,
                color='orange',
                alpha=1,
            )

            ax.text((self.TimeArray.time_array[self.ObservationsTwilight.complete[0]]+30*u.min).to_value(
                format='datetime'),
                15,
                'Observation (with twilight): ' +
                self.TimeArray.time_array[self.ObservationsTwilight.complete[0]].to_value(
                format='datetime').strftime('%H:%M') + ' - ' +
                self.TimeArray.time_array[self.ObservationsTwilight.complete[-1]].to_value(
                format='datetime').strftime('%H:%M'),
                fontsize=10,
            )

            ax.axvline(self.TimeArray.time_array[self.ObservationsTwilight.complete[0]].to_value(format='datetime'),
                       ls='--',
                       lw=2,
                       color='orange'
                       )
            ax.axvline(self.TimeArray.time_array[self.ObservationsTwilight.complete[-1]].to_value(format='datetime'),
                       ls='--',
                       lw=2,
                       color='orange')


        ax.axvline(self.TimeArray.T1.to_value(format='datetime'),
                   ls='--', lw=2, color='black')
        ax.axvline(self.TimeArray.T4.to_value(format='datetime'),
                   ls='--', lw=2, color='black')
        if len(self.ObservationsFull.complete) != 0:
            ax.axvline(self.TimeArray.time_array[self.ObservationsFull.complete[0]].to_value(
                format='datetime'), ls='--', lw=2, color='lime')
            ax.axvline(self.TimeArray.time_array[self.ObservationsFull.complete[-1]].to_value(
                format='datetime'), ls='--', lw=2, color='lime')

        return

    def _write_system_parameters(self,
                                 ax: plt.Axes):
        """
        Write the system parameters for given system, including Right Ascension, Declination,V magnitude, Moon angle separation, Transit length, Period, ingress and egress timings.

        Parameters
        ----------
        ax : plt.Axes
            DESCRIPTION.

        Returns
        -------
        None.

        """
        from matplotlib import rc

        if self.AltitudeArray.MoonSeparation.value > 45:
            color_moon = 'white'
        elif self.AltitudeArray.MoonSeparation.value > 30:
            color_moon = 'orange'
        else:
            color_moon = 'red'

        format_text = {'style': 'italic',
                       'size': 12
                       }

        text = ax.text(self.TimeArray.Sunset, 97,
                       f"RA: {self.row['Position.RightAscension']:.2f} ",
                       **format_text
                       )

        text = ax.annotate(
            f"DEC: {self.row['Position.Declination']:.2f} ",
            xycoords=text,
            xy=(1, 0), verticalalignment="bottom",
            **format_text
        )

        text = ax.annotate(
            f"; V = {self.row['Magnitude.V']:.3f} ",
            xycoords=text,
            xy=(1, 0), verticalalignment="bottom",
            **format_text
        )
        text = ax.annotate(
            f"; Moon = {self.AltitudeArray.MoonSeparation.value:.0f} deg ",
            xycoords=text,
            xy=(1, 0), verticalalignment="bottom",
            bbox = dict(facecolor=color_moon, edgecolor=color_moon,
                    boxstyle='round,pad=1'),
            **format_text,
        )
        text = ax.annotate(
            '; $T_{14}$' + f" = {self.row['Planet.TransitDuration']:.3f} h ",
            xycoords=text,
            xy=(1, 0), verticalalignment="bottom",
            **format_text
        )

        text = ax.annotate(
            f"; P = {self.row['Planet.Period']:.5f} d ",
            xycoords=text,
            xy=(1, 0), verticalalignment="bottom",
            **format_text
        )
        
        if 'Flag.TransitTimingVariations' in self.row.keys():
            if self.row['Flag.TransitTimingVariations']:
                text = ax.annotate(
                    f"; TTV detected",
                    xycoords=text,
                    xy=(1, 0), verticalalignment="bottom",
                    bbox = dict(facecolor='red', edgecolor='red',
                    boxstyle='round,pad=1'),
                    **format_text
                )
        
        sunset_text = ax.text(self.TimeArray.Sunset, -10,
                              f"ingress= {self.TimeArray.T1.to_value('datetime').strftime('%H:%M')}",
                              style='italic',
                              bbox=dict(
                                  facecolor='none',
                                  edgecolor='black',
                                  pad=7,
                    boxstyle='round,pad=1'
                              )
                              )

        sunset_text = ax.annotate(
            (f"Observations (no twilight): {self.TimeArray.time_array[self.ObservationsFull.complete[0]].to_value(format='datetime').strftime('%H:%M')} " +
             f"- {self.TimeArray.time_array[self.ObservationsFull.complete[-1]].to_value(format='datetime').strftime('%H:%M')}"),
            xycoords=sunset_text,
            xy=(1.4, 0), verticalalignment="bottom",
            bbox=dict(
                facecolor='green',
                edgecolor='black',
                boxstyle='round, pad=1'
            )
        )

        if not(self.full_observation_time == self.twilight_observation_time):
            sunset_text = ax.annotate(
                (f"Observation (with twilight): {self.TimeArray.time_array[self.ObservationsTwilight.complete[0]].to_value(format='datetime').strftime('%H:%M')} " +
                f"- {self.TimeArray.time_array[self.ObservationsTwilight.complete[-1]].to_value(format='datetime').strftime('%H:%M')}"),
                xycoords=sunset_text,
                xy=(1.4, 0), verticalalignment="bottom",
                bbox=dict(
                    facecolor='orange',
                    edgecolor='black',
                    boxstyle='round,pad=1'
                ),
            )

        ax.text(self.TimeArray.Sunrise, -10,
                f"egress= {self.TimeArray.T4.to_value('datetime').strftime('%H:%M')}",
                style='italic',
                bbox=dict(
                    facecolor='none',
                    edgecolor='black',
                    pad=7)
                )
        return
